package loderunner;

import java.io.*;
import java.net.*;
import java.awt.*;
import java.util.*;

class LevelFormatException extends IOException {}
/*
 *        .
 */
public class Level {
    
    // 
    static int EMPTY = 0;
    static int BRICK1 = 1;
    static int BRICK2 = 2;	
    static int STAIRS = 3;	
    static int EXITSTAIRS = 5;	
    static int ROW = 4;
    static int LODE = 6;        
    private static int RUNNER = 7;    
    private static int ENEMY = 8;
    private static int TRAP = 9;
    static int HOLE = 10;
    static int MANINHOLE = 11;
    
    // 
    private static int width = 32;
    private static int height = 22;
    private static int cellSize = 16;		
    private static int [] coreCodes = {1,2,3,4};
    private byte [][] map, sourceMap, grayMap;
    private final URL fileBase;	
    private int boxesNumber, enemysNumber;
    Runner runner;
    Enemy [] enemys;
    private Vector exitStairs;
    Vector enemyList;
    public Vector holes;
    public Vector lodes;
    private Vector traps;	    
    Image coreImages[];
    Image rImages[], eImages[];    
    private ArrayList gray;
    private byte grayINF = 127;
    
	
    Level(URL fileBase,Image images[],Image [] rImages, Image [] eImages){
        this.fileBase = fileBase;
        map = new byte [height] [width];
        sourceMap = new byte [height] [width];
        grayMap = new byte [height] [width];

        coreImages = images;
        this.rImages = rImages;
        this.eImages = eImages;
        holes = new Vector ();
        lodes = new Vector();
        exitStairs = new Vector();
        traps = new Vector();
    } 
	
    //  
    void loadLevel(String fileName) throws IOException,LevelFormatException {		
        URL url = new URL(fileBase + fileName);		
        URLConnection urlConnection = url.openConnection();
        InputStream iStream = urlConnection.getInputStream();
        BufferedReader bFile = new BufferedReader(new InputStreamReader(iStream));
        try {							
            String str; 			
            for (int i =0; i<height; i++){
                str = bFile.readLine();				
                if ((str ==  null) || (str.length () != width )) throw new LevelFormatException();
                for (int j = 0; j < width;  j++)
                sourceMap[i][j] = Byte.parseByte(str.substring (j,j+1));
            }			
        } catch (IOException e) {
            if (bFile != null) bFile.close();
            if (iStream != null) iStream.close();
            throw e;
        }		
        reInitLevel();
    }
	
    //     
    public void reInitLevel(){
        lodes.clear();
        holes.clear();
        exitStairs.clear();
        traps.clear();
        enemysNumber = 0;
        boxesNumber = 0;
        
        for (int i =0; i<height; i++)
            for (int j = 0; j < width;  j++) 
                map[i][j] = sourceMap[i][j];
        for (int i =0; i<height; i++)
            for (int j = 0; j < width;  j++) {			
                boxesNumber += (map[i][j] == LODE) ? 1 : 0;
                if (map[i][j] == LODE) {
                    lodes.add(new Lode(j,i));
                    map[i][j] = (byte)EMPTY;
                }		
                if (map[i][j] == TRAP) { traps.add(new Byte ((byte)i)); traps.add(new Byte((byte)j)); map[i][j] = (byte)EMPTY;}
                enemysNumber += (map[i][j] == ENEMY) ? 1 : 0;
                if (map[i][j] == EXITSTAIRS) {
                    exitStairs.add(new Byte((byte)i));
                    exitStairs.add(new Byte((byte)j));
                    map[i][j] = (byte)EMPTY;
                }
            }		
        try {
            createMans();
            buildRoutes(runner.getPos().getX(),runner.getPos().getY());
        } catch (LevelFormatException e){			
        }       
        
        try {
            Thread.currentThread().wait();
        } catch (Exception e) {            
        }
    }

    // 
    private void createMans() throws LevelFormatException {
        enemyList = new Vector ();
        for (int i = 0; i < height; i++)
            for (int j = 0; j < width; j++) {
                if (map[i][j] == ENEMY)	{ 
                    enemyList.addElement(new Enemy (eImages,this,j,i));  
                    map[i][j] = (byte)EMPTY;
                }

                if (map[i][j] == RUNNER) 
                    if (runner == null)
                        runner = new Runner (rImages,this,j,i);
                    else runner.setPos(new Position(j,i));                   
            }
        
        if (runner == null) throw new LevelFormatException();
        map[runner.pos.getY()][runner.pos.getX()] = (byte)EMPTY;    

        for (int i = 0; i < enemyList.size(); i++) 
            ((Enemy) enemyList.get(i)).setHeroe(runner);
    }

    //   
    public Man [] getMen(){
        Man [] men = new Man [1+enemyList.size()];
        men[0] = runner;
        for (int i = 0; i < enemyList.size();i++)
            men[i+1] = (Enemy) enemyList.elementAt(i); 
        return men;
    }

    // 
    public void paint(Graphics g){
        for (int i = 0; i < height; i++)
            for (int j = 0; j < width; j++) 
                if ((map[i][j] >= 1) && (map[i][j] <= 4))
                    g.drawImage(coreImages [map[i][j]-1], cellSize*j, cellSize*i, null);
        for (int i = 0; i < traps.size(); i += 2) {
            g.drawImage(coreImages [0], cellSize * ((Byte)traps.get(i + 1)).byteValue(), cellSize * ((Byte)traps.get(i)).byteValue(), null);
        }
    }

    // 
    public void paintHoles(Graphics g){
        for (int i = 0; i < holes.size(); i++)
            ((Hole)holes.get(i)).draw(g);
    }

    // 
    public void paintLodes(Graphics g){
        for (int i = 0; i < lodes.size(); i++)
            ((Lode)lodes.get(i)).draw(g);
    }

    //  
    public int getBoxesNumber() {return boxesNumber; }

    //  
    public int getEnemysNumber() {return enemysNumber; }

    //  
    public static int getCellSize(){
        return cellSize;
    }

    //    
    public int getWidth(){
        return width;
    }

    //    
    public int getHeight(){
        return height;
    }

    //     (x,y)   
    public int getBoxAT(int y, int x){
        if ((x>=width) || (x<0) || (y<0) || (y >= height)) return BRICK2;
        else return map[y][x];		
    }

    //     (x,y)  ,  
    public int getSBoxAT(int y, int x){
        if ((x>=width) || (x<0) || (y<0) || (y >= height)) return BRICK2;
        else if ((sourceMap[y][x] != LODE) && (sourceMap[y][x] != RUNNER) && (sourceMap[y][x] != ENEMY) && (sourceMap[y][x] != EXITSTAIRS)) return sourceMap[y][x]; 
            else return map[y][x];
    }

    // ,   (x,y),    
    public byte getEvrAT(int y, int x){
        if ((x>=width) || (x<0) || (y<0) || (y >= height)) return grayINF;
        else return grayMap[y][x];		
    }

    //  
    public void newHole(int x, int y){
        boolean f =false;		
        for (int i =0; i < holes.size(); i++) if (((Hole)holes.get(i)).isEqual(x,y)) {f = true; break;}
        if (!f)	holes.add(new Hole(x,y,this));
    }
	
	//   (x,y)
    public void setBoxAT(int y, int x, int  type){
        map[y][x] = (byte)type;
    }

    //  
    public void showExit(){
        for (int i=0; i < exitStairs.size(); i=i+2){
            map[((Byte)exitStairs.get(i)).byteValue()][((Byte)exitStairs.get(i+1)).byteValue()] = (byte)STAIRS;
        }			
    }

    //       
    public boolean isExit(int y, int x){
        int l = exitStairs.size ();
        if ((y == 0) && (getBoxAT(y,x) == STAIRS)) return true;
        else return false;				
    }

    //  ,     
    public void buildRoutes(int x, int y) {
        gray = new ArrayList();
        gray.add(new Integer(y * 32 + x));
        for (int i = 0;  i < height; i++)            
            for (int j = 0; j < width; j++) 
                grayMap[i][j] = grayINF;        
        
        grayMap[y][x] = 0;
        while (gray.size() > 0) {
            Object o = gray.remove(0);
            if (o != null) {
                y = ((Integer) o).intValue() / 32;
                x = ((Integer) o).intValue() % 32;                            

                if ((getSBoxAT(y - 1, x) != BRICK1) && (getSBoxAT(y - 1, x) != BRICK2) && (grayMap[y - 1][x] == grayINF)) {
                    gray.add(new Integer( (y - 1) * 32 + x));
                    grayMap[y - 1][x] = (byte)(grayMap[y][x] + 1);
                }
                
                if  (((getSBoxAT(y, x + 1) == STAIRS) || (getSBoxAT(y, x + 1) == ROW) || (((getSBoxAT(y + 1, x + 1) == BRICK1) || (getSBoxAT(y + 1, x + 1) == BRICK2)||(getSBoxAT(y + 1, x + 1) == STAIRS)) && (getSBoxAT(y, x + 1) == EMPTY)) ) && (grayMap[y][x + 1] == grayINF)) {
                    gray.add(new Integer( y * 32 + x + 1));
                    grayMap[y][x + 1] = (byte)(grayMap[y][x] + 1);
                }            

                if ((getSBoxAT(y + 1, x) == STAIRS) && (grayMap[y + 1][x] == grayINF)) {
                    gray.add(new Integer( (y + 1) * 32 + x));
                    grayMap[y + 1][x] = (byte)(grayMap[y][x] + 1);
                }

                if  (((getSBoxAT(y, x - 1) == STAIRS) || (getSBoxAT(y, x - 1) == ROW) || (((getSBoxAT(y + 1, x - 1) == BRICK1) || (getSBoxAT(y + 1, x - 1) == BRICK2) || (getSBoxAT(y + 1, x - 1) == STAIRS)) && (getSBoxAT(y, x - 1) == EMPTY))) && (grayMap[y][x - 1] == grayINF))  {
                    gray.add(new Integer( y * 32 + x - 1));
                    grayMap[y][x - 1] = (byte)(grayMap[y][x] + 1);
                }            
            }
        }
    }
    
    //  ;
    public byte getINF() {
        return grayINF;
    }
}